from .hash_table import HashTable
from .pq import DistHeap
from .util import NNeighbor, euclidean_dist

class LSHRandom:
    def __init__(self, data, hash_len, num_tables = 2):
        self.n_rows, self.dims = data.shape
        self.hash_len = hash_len
        self.tables = []
        for i in range(num_tables):
            t = HashTable(hash_len, self.dims)
            self.tables.append(t)
            self.tables[i].build_table(data)

    def add_new_entry(self, data_point):
        for t in self.tables:
            t.add_new_entry(data_point, self.n_rows)

    def __get_distances(self, candidate, candidate_point, query_point, dist_func):
        return NNeighbor(candidate, dist_func(candidate_point, query_point), candidate_point)

    def knn_search(self, data, query_point, k, buckets, dist_func = euclidean_dist, radius = 2):
        best_candidates = set()
        for t in self.tables:
            tbc = t.knn_search(query_point, k, buckets, radius)
            best_candidates.update(tbc)
        nn_heap = DistHeap()
        for c in best_candidates:
            candidate_point = data[c, :]
            nn = self.__get_distances(c, candidate_point, query_point, dist_func)
            nn_heap.push(nn)
        return nn_heap[:k]





from lshashing import LSHRandom
import numpy as np

def test_lshashing():
    sample_data = np.random.randint(size = (20, 20), low = 0, high = 100)
    point = np.random.randint(size = (1, 20), low = 0, high = 100)

    lsh_random = LSHRandom(sample_data, 4)
    assert lsh_random.knn_search(sample_data, point[0], 4, 3)



from .util import generate_rand_proj, hash_fun
from .hash_util import *

class HashTable:
    def __init__(self, hash_len, dim):
        self.hash_table = {}
        self.hash_len = hash_len
        self.rand_proj = generate_rand_proj(self.hash_len, dim)

    def build_table(self, data):
        data_hashes = hash_fun(data, self.rand_proj)
        return fill_table(self.hash_table, data_hashes)

    def add_new_entry(self, new_entry, n_rows):
        new_hash = hash_fun(new_entry, self.rand_proj)
        return fill_table(self.hash_table, new_hash, n_rows)

    def knn_search(self, query_point, k, buckets, radius = 2):
        if buckets > self.hash_len:
            buckets = self.hash_len - 1
        new_hash = hash_fun(query_point, self.rand_proj)
        candidates = set()
        for h in new_hash:
            if h in self.hash_table:
                candidates.update(self.hash_table[h])

            near_candidates = search_near_hashes(self.hash_table, h,
                                                 buckets, len(candidates), k, radius)
            candidates.update(near_candidates)
        return candidates

    #def __repr__(self):
    #    return self.hash_table




from setuptools import setup

with open("README.md", "r", encoding="utf8") as rm:
    readme = rm.read()
    
with open("requirements.txt") as rq:
    requirements = rq.read().split('\n')

setup(
      name="lshashing",
      version="0.0.1", 
      description="Nearest neighbors search using locality-sensitive hashing",
      packages=["lshashing"],
      install_requires=requirements,
      long_description=readme,
      long_description_content_type="text/markdown",
      url="https://github.com/MNoorFawi/lshashing",
      author="Muhammad N. Fawi",
      author_email="m.noor.fawi@gmail.com",
      license="MIT",
      classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ]
      )
